;;
;; Low-level WebDAV client implementation.
;;
;; Based on the JavaScript WebDAV client by Guido Wesdorp.
;; Copyright 2008 Ivan Raikov.
;;
;; This program is free software: you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation, either version 3 of the
;; License, or (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.
;;
;; A full copy of the GPL license can be found at
;; <http://www.gnu.org/licenses/>.
;;


(define-extension webdav-client)

(require-extension  srfi-1)
(require-extension  syntax-case)
(require-extension  utf8)
(require-extension  utf8-srfi-13)
(require-extension  utf8-srfi-14)
(require-extension  stx-engine)
(require-extension  matchable)
(require-extension  datatype)
(require-extension  sxpath-plus)
(require-extension  sxml-transforms)
(require-extension  sxml-tools)
(require-extension  http-client)
(require-extension  http-utils)
(require-extension  http-auth)


(declare (export webdav:get
		 webdav:put
		 webdav:delete
		 webdav:copy
		 webdav:move
		 webdav:mkcol
		 webdav:propfind
		 webdav:proppatch
		 webdav:lock
		 webdav:unlock
		 webdav:options
		 webdav:report
		 webdav:status-codes))


(define (lookup-def k lst . rest)
  (let-optionals rest ((default #f))
      (let ((kv (assoc k lst)))
	(if (not kv) default
	    (match kv ((k v) v) (else (cdr kv)))))))


;;; Procedures for string concatenation and pretty-printing

(define (s+ . lst)    (string-concatenate (map ->string lst)))
(define (sw+ lst)     (string-intersperse (filter-map (lambda (x) (and x (->string x))) lst) " "))
(define (s\ p . lst)  (string-intersperse (map ->string lst) p))
(define (sl\ p lst)   (string-intersperse (map ->string lst) p))
(define nl "\n")

;;; Error/warning procedures for the XML parser

(define (parser-error port message . specialising-msgs)
  (error (string-append message (string-concatenate (map ->string specialising-msgs)))))

(define (ssax:warn port message . specialising-msgs)
  (print-error-message message (current-output-port) "Warning")
  (print (string-concatenate (map ->string specialising-msgs))))


(include "SXML.scm")
(include "SSAX.scm")
(include "SXML-to-XML.scm")

(define webdav:status-codes
  `((100   . "Continue")
    (101   . "Switching Protocols")
    (102   . "Processing")
    (200   . "OK")
    (201   . "Created")
    (202   . "Accepted")
    (203   . "None-Authoritive Information")
    (204   . "No Content")
    (1223  . "No Content")
    (205 . "Reset Content")
    (206 . "Partial Content")
    (207 . "Multi-Status")
    (300 . "Multiple Choices")
    (301 . "Moved Permanently")
    (302 . "Found")
    (303 . "See Other")
    (304 . "Not Modified")
    (305 . "Use Proxy")
    (307 . "Redirect")
    (400 . "Bad Request")
    (401 . "Unauthorized")
    (402 . "Payment Required")
    (403 . "Forbidden")
    (404 . "Not Found")
    (405 . "Method Not Allowed")
    (406 . "Not Acceptable")
    (407 . "Proxy Authentication Required")
    (408 . "Request Time-out")
    (409 . "Conflict")
    (410 . "Gone")
    (411 . "Length Required")
    (412 . "Precondition Failed")
    (413 . "Request Entity Too Large")
    (414 . "Request-URI Too Large")
    (415 . "Unsupported Media Type")
    (416 . "Requested range not satisfiable")
    (417 . "Expectation Failed")
    (422 . "Unprocessable Entity")
    (423 . "Locked")
    (424 . "Failed Dependency")
    (500 . "Internal Server Error")
    (501 . "Not Implemented")
    (502 . "Bad Gateway")
    (503 . "Service Unavailable")
    (504 . "Gateway Time-out")
    (505 . "HTTP Version not supported")
    (507 . "Insufficient Storage")))


(define (add-locktoken locktoken attrs)
  (if locktoken (cons `("If" . ,(string-append "<" (->string locktoken) ">")) attrs) attrs))

(define (make-request method path . rest)
  (match path
	 ((auth path)  
	  (let* ((attrs (if (pair? rest) (car rest) (list)))
		 (rest  (if (pair? rest) (cdr rest) rest))
		 (req   (apply http:make-request (cons* method path attrs rest))))
	    (lambda (h) (http:authenticate h auth req #t))))
	 (else (apply http:make-request (cons* method path rest)))))
  
(define (webdav:options path . rest)  
  (let-optionals rest ((handler default-handler))
    (let ((req  (make-request 'OPTIONS path)))
      (response-wrapper handler req))))

(define (webdav:get path . rest)  
  (let-optionals rest ((handler default-handler))
    (let ((req  (make-request 'GET path)))
      (response-wrapper handler req))))


(define (webdav:put path content . rest)
  (let-optionals rest ((locktoken #f) (handler default-handler))
    (let* ((attrs `(("Content-type" . "text/xml,charset=UTF-8")))
	   (attrs  (add-locktoken locktoken attrs))
	   (req    (make-request 'PUT path attrs content)))
      (response-wrapper handler req))))


(define (webdav:delete path . rest)
  (let-optionals rest ((handler default-handler))
    (let ((req (make-request 'DELETE path)))
      (response-wrapper handler req))))


(define (webdav:copy from-path to-path . rest)
  (let-optionals rest ((overwrite #f) (locktoken #f) (handler default-handler))
    (let* ((attrs `(("Destination" . ,(->string to-path))))
	   (attrs  (if overwrite (cons `("Overwrite" . "F") attrs) attrs))
	   (attrs  (add-locktoken locktoken attrs))
	   (req    (make-request 'COPY from-path attrs)))
      (response-wrapper handler req))))
      


(define (webdav:move from-path to-path . rest)
  (let-optionals rest ((overwrite #f) (locktoken #f) (handler default-handler))
    (let* ((attrs `(("Destination" . ,(->string to-path))))
	   (attrs  (if overwrite (cons `("Overwrite" . "F") attrs) attrs))
	   (attrs  (add-locktoken locktoken attrs))
	   (req    (make-request 'MOVE from-path attrs)))
      (response-wrapper handler req))))


(define (webdav:mkcol path . rest)
  (let-optionals rest ((locktoken #f) (handler default-handler))
    (let* ((attrs `())
	   (attrs  (add-locktoken locktoken attrs))
	   (req    (make-request 'MKCOL path attrs)))
      (response-wrapper handler req))))


(define (namespace-list namespaces)
  (map (lambda (x) `(,(string->symbol (s+ "xmlns:" (first x))) ,(second x))) namespaces))

(define (webdav:propfind path . rest)
  (let-optionals rest ((props (list)) (namespaces (list)) (depth 0) (handler default-handler))
    (let* ((attrs `(("Depth" . ,(->string (if (positive? depth) depth 0))) 
		    ("Content-type" . "text/xml; charset=UTF-8")))
	   (content `(D:propfind (@ (xmlns:D "DAV:") . ,(namespace-list namespaces) )
				 ,(if (null? props) `(D:allprop )
				      `(D:prop . ,(map (lambda (x) `(,(string->symbol (s+ (first x) ":" (second x)))))
						       props)))))
	   (req     (make-request 'PROPFIND path attrs (generate-XML `(begin ,content)))))
      (response-wrapper handler req))))


(define (webdav:report path content . rest)
  (let-optionals rest ((depth 0) (handler default-handler))
    (let* ((attrs `(("Content-type" . "text/xml; charset=UTF-8")
		    ("Depth" . ,(->string (if (positive? depth) depth 0)))))
	   (req     (make-request 'REPORT path attrs (generate-XML `(begin ,content)))))
      (response-wrapper handler req))))


(define (webdav:proppatch path  . rest)
  (let-optionals rest ((setprops (list)) (delprops (list)) (namespaces (list)) 
		       (locktoken #f)    (handler default-handler))
    (let* ((attrs `(("Content-type" . "text/xml; charset=UTF-8")))
	   (attrs  (add-locktoken locktoken attrs))
	   (content `(D:propertyupdate
		      (@ (xmlns:D "DAV:")  . ,(namespace-list namespaces))
		      (D:set (D:prop .,(map (lambda (x) `(,x)) setprops)))
		      (D:remove (D:prop . ,(map (lambda (x) `(,x)) delprops)))))
	   (req     (make-request 'PROPPATCH path attrs (generate-XML `(begin ,content)))))
      (response-wrapper handler req))))


(define (webdav:lock path owner . rest)
  (let-optionals rest ((scope 'exclusive) (depth #f) (timeout "Infinite, Second-4100000000") 
		       (locktoken #f) (handler default-handler) )
    (let* ((attrs   `(("Content-type" . "text/xml; charset=UTF-8")))
	   (attrs   (if depth (cons `("Depth" . ,(->string depth)) attrs) attrs))
	   (attrs   (if (integer? timeout) 
			(cons `("Timeout" . ,(string-append "Second-" (->string timeout))) attrs) 
			(cons `("Timeout" . ,(->string timeout)) attrs)))
	   (attrs   (add-locktoken locktoken attrs))
	   (content `(D:lockinfo (@ (xmlns:D "DAV:"))
				 (D:lockscope (,(string->symbol (s+ "D:"  scope))))
				 (D:locktype  (D:write))
				 (D:owner     (D:href ,owner))))
	   (req     (make-request 'LOCK path attrs (generate-XML `(begin ,content)))))
      (response-wrapper handler req))))



(define (webdav:unlock path locktoken . rest)
  (let-optionals rest ((handler default-handler))
   (let* ((attrs `(("Lock-Token" . ,(string-append "<" (->string locktoken) ">"))))
	  (req     (make-request 'UNLOCK path attrs)))
     (response-wrapper handler req))))


;;
;; Handles content parsing and calls the given handler when done
;;
(define (response-wrapper handler req) 
  (let-values (((status headers in out) 
		(if (procedure? req) (req (lambda (fst headers in out . rest) (values fst headers in out)))
		    (http:send-request req))))
      (let* ((content-type ((lambda (x) (and x (string->symbol (car (string-split x ";")))))
			    (alist-ref "content-type" headers string-ci=?)))
	     (content (case content-type 
			((text/xml) (ssax:xml->sxml in '()))
			(else (read-lines in)))))
	(handler status content headers in out))))


(define (default-handler status content headers in out . rest)
  (close-input-port in)
  (close-output-port out)
  `((status ,status)  (status-message ,(lookup-def status webdav:status-codes))
    (headers . ,headers) (content . ,content)))
